﻿Shader "Custom/Ray Marching/Ray Marching Instanced"
{
	Properties
	{
		_MipLevel("Mip Level", Int) = 0
		_GridSize("Grid Size", Int) = 0
		_VolumeTex("Volume Texture", 3D) = "" {}
		_NumRayStepMax("_Num Ray Step Max", Int) = 32
		_Threshold("Intensity Threshold", Range(0, 1)) = 0.5
	}

	CGINCLUDE

	#include "UnityCG.cginc"
	#pragma target 5.0

	int _MipLevel;
	int _NumRayStepMax;
	float _Threshold;
	float4 _TextureSize;
	sampler3D _VolumeTex;

	uniform	StructuredBuffer<int> _CubeIndices;
	uniform	StructuredBuffer<float3> _CubeVertices;
	uniform	StructuredBuffer<float4x4> _CubeMatrices;
	
	struct v2f
	{		
		float3 rayDir : C0LOR0;
		float3 rayEndPos : C0LOR1;
		float4 pos : SV_POSITION;
		float4x4 modelMatrix : world;
	};

	float sampleVolumeMip(float3 p, int mipLevel)
	{
		return tex3Dlod(_VolumeTex, float4(p, mipLevel)).a;
	}

	float3 getNormal(float3 position, float3 dataStep)
	{
		float dx = sampleVolumeMip(position + float3(dataStep.x, 0, 0), _MipLevel) - sampleVolumeMip(position + float3(-dataStep.y, 0, 0), _MipLevel);
		float dy = sampleVolumeMip(position + float3(0, dataStep.y, 0), _MipLevel) - sampleVolumeMip(position + float3(0, -dataStep.y, 0), _MipLevel);
		float dz = sampleVolumeMip(position + float3(0, 0, dataStep.z), _MipLevel) - sampleVolumeMip(position + float3(0, 0, -dataStep.z), _MipLevel);

		return normalize(float3(dx, dy, dz));
	}

	v2f vert(uint id : SV_VertexID, uint instanceId : SV_InstanceID)
	{
		float4x4 modelMatrix = _CubeMatrices[instanceId];
		float4x4 modelMatrixInv = transpose(modelMatrix);

		float3 objectPos = _CubeVertices[_CubeIndices[id]];
		float4 worldPos = mul(modelMatrix, float4(objectPos, 1));		

		v2f output;
		output.rayEndPos = objectPos;
		output.modelMatrix = modelMatrix;
		output.pos = mul(UNITY_MATRIX_VP, worldPos);
		output.rayDir = mul(modelMatrixInv, worldPos - _WorldSpaceCameraPos);
		return output;
	}
	
	#define STEP_CNT 128
	#define STEP_SIZE 1 / 64

	float4 frag_surf_opaque(v2f input) : COLOR0
	{		
		float3 rayDir = normalize(input.rayDir);
		float3 rayStep = rayDir * STEP_SIZE;

		float3 rayEndPos = input.rayEndPos;
		float3 t = max((0.5 - rayEndPos) / -rayDir, (-0.5 - rayEndPos) / -rayDir);
		float3 rayStartPos = rayEndPos + (min(t.x, min(t.y, t.z)) * -rayDir);

		rayEndPos += 0.5;
		rayStartPos += 0.5;
		
		float intensity = 0;
		float threshold = _Threshold;
		float3 rayCurrentPos = rayStartPos;

		// Linear search
		for (int k = 0; k < STEP_CNT; k++)
		{
			intensity = sampleVolumeMip(rayCurrentPos, _MipLevel);
			if (intensity > threshold) break;
			rayCurrentPos += rayStep;
		}

		if (intensity < threshold) discard;

		rayStep *= 0.5;
		rayCurrentPos -= rayStep;

		// Binary search
		for (uint j = 0; j < 4; j++)
		{
			rayStep *= 0.5;
			intensity = sampleVolumeMip(rayCurrentPos, _MipLevel);
			rayCurrentPos += (intensity >= threshold) ? -rayStep : rayStep;
		}

		float3 texelSize = 1.0 / _TextureSize;
		float3 normal = getNormal(rayCurrentPos, texelSize);
		float ndotl = pow(max(0.0, dot(normal, rayDir)), 1);

		return float4(1, 1, 1, 1) * ndotl;
	}

	ENDCG

		Subshader
		{
			ZTest LEqual
			ZWrite On
			Cull Front

			Pass
			{
				CGPROGRAM
				#pragma vertex vert		
				#pragma fragment frag_surf_opaque		
				ENDCG
			}
		}

		Fallback off
}
